from uuid import UUID

from django.conf import settings
from rest_framework import serializers
from rest_framework.exceptions import ValidationError

from arkindex.dataimport.models import ActivityState, DataFile, DataImport, DataImportMode, WorkerRun
from arkindex.dataimport.serializers.git import RevisionSerializer
from arkindex.dataimport.serializers.workers import WorkerLightSerializer
from arkindex.documents.models import Corpus, Element, ElementType
from arkindex.documents.serializers.elements import ElementSlimSerializer
from arkindex.project.serializer_fields import BestClassField, EnumField, LinearRingField
from arkindex.users.models import Role
from arkindex.users.utils import get_max_level
from ponos.models import Farm, State
from transkribus import TranskribusAPI


class DataImportLightSerializer(serializers.ModelSerializer):
    """
    Serialize a data importing workflow
    """

    state = EnumField(State)
    mode = EnumField(DataImportMode, read_only=True)
    creator = serializers.HiddenField(default=serializers.CurrentUserDefault())
    workflow = serializers.HyperlinkedRelatedField(read_only=True, view_name='ponos:workflow-details')

    class Meta:
        model = DataImport
        fields = (
            'name',
            'id',
            'state',
            'mode',
            'corpus',
            'creator',
            'workflow',
        )
        read_only_fields = ('id', 'state', 'mode', 'corpus', 'creator', 'workflow')


class DataImportSerializer(DataImportLightSerializer):
    """
    Serialize a data importing workflow with its settings
    """

    # Redefine state as writable
    state = EnumField(State, read_only=True)
    revision = RevisionSerializer(read_only=True)
    element = ElementSlimSerializer(read_only=True)
    folder_type = serializers.SlugField(source='folder_type.slug', default=None, read_only=True)
    element_type = serializers.SlugRelatedField(queryset=ElementType.objects.none(), slug_field='slug', allow_null=True)
    element_name_contains = serializers.CharField(
        source='name_contains',
        allow_null=True,
        allow_blank=True,
        max_length=250
    )
    activity_state = EnumField(ActivityState, read_only=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        dataimport = self.context.get('dataimport')
        if not dataimport or not dataimport.corpus:
            return
        self.fields['element_type'].queryset = ElementType.objects.filter(corpus=dataimport.corpus)

    class Meta(DataImportLightSerializer.Meta):
        fields = DataImportLightSerializer.Meta.fields + (
            'files',
            'revision',
            'element',
            'folder_type',
            'element_type',
            'element_name_contains',
            'load_children',
            'use_cache',
            'activity_state',
        )
        read_only_fields = DataImportLightSerializer.Meta.read_only_fields + (
            'files',
            'revision',
            'element',
            'folder_type',
            'use_cache',
            'activity_state',
        )

    def validate(self, data):
        data = super().validate(data)
        # Editing a dataimport name only is always allowed
        if set(data.keys()) == {'name'}:
            return data

        if not self.instance:
            return
        if self.instance.state == State.Running:
            raise serializers.ValidationError({'__all__': ['Cannot edit a workflow while it is running']})
        if self.instance.mode != DataImportMode.Workers:
            raise serializers.ValidationError({'__all__': [f'Only processes of mode {DataImportMode.Workers} can be updated']})
        return data


class DataImportFromFilesSerializer(serializers.Serializer):

    mode = EnumField(DataImportMode, default=DataImportMode.Images)
    files = serializers.PrimaryKeyRelatedField(queryset=DataFile.objects.all(), many=True)
    folder_id = serializers.UUIDField(required=False, allow_null=True)
    folder_type = serializers.SlugField(required=False, allow_null=True)
    element_type = serializers.SlugField()

    default_error_messages = {
        'mode_not_allowed': 'This mode is not allowed when importing from files',
        'files_required': 'At least one file is required to start an import',
        'unique_corpus': 'Imports can only run on files from a single corpus',
        'corpus_read_only': 'Cannot write in corpus',
        'folder_not_found': 'Folder does not exist',
        'unique_pdf': 'Cannot import more than one PDF at once',
        'pdf_only': 'PDF imports can only import PDF files',
        'images_only': 'Image imports can only import image files',
        'iiif_only': 'IIIF imports can only import IIIF documents',
        'folder_required': 'Either folder_type, folder_id or both are required',
        'iiif_folder_required': 'IIIF imports require both folder_type and element_type',
        'type_not_found': 'Element type {slug!r} does not exist',
        'type_not_folder': 'Element type {slug!r} is not a folder',
        'type_folder': 'Element type {slug!r} is a folder',
    }

    def validate_mode(self, mode):
        if mode not in (DataImportMode.Images, DataImportMode.PDF, DataImportMode.IIIF):
            self.fail('mode_not_allowed')
        return mode

    def validate_files(self, files):
        if not files:
            self.fail('files_required')
        corpora = set(f.corpus for f in files)
        if len(corpora) > 1:
            self.fail('unique_corpus')
        corpus = corpora.pop()
        if corpus not in Corpus.objects.writable(self.context['request'].user):
            self.fail('corpus_read_only')
        return files

    def validate_folder_id(self, folder_id):
        if not folder_id:
            return None
        try:
            return Element.objects.get(
                id=folder_id,
                type__folder=True,
            )
        except Element.DoesNotExist:
            self.fail('folder_not_found')

    def validate(self, data):
        if data['mode'] == DataImportMode.PDF:
            if not all(f.content_type == 'application/pdf' for f in data['files']):
                self.fail('pdf_only')

        elif data['mode'] == DataImportMode.Images:
            if not all(f.content_type.startswith('image/') for f in data['files']):
                self.fail('images_only')

        elif data['mode'] == DataImportMode.IIIF:
            if not set(f.content_type.split(';')[0] for f in data['files']) <= {'application/json', 'application/ld+json'}:
                self.fail('iiif_only')

        else:
            raise NotImplementedError

        element_type_slug = data.get('element_type')
        element_type = data['files'][0].corpus.types.filter(slug=element_type_slug).first()
        if not element_type:
            self.fail('type_not_found', slug=element_type_slug)
        if element_type.folder:
            self.fail('type_folder', slug=element_type_slug)

        data['element_type'] = element_type
        folder_type_slug, folder_id = data.get('folder_type'), data.get('folder_id')

        if folder_type_slug:
            folder_type = data['files'][0].corpus.types.filter(slug=folder_type_slug).first()
            if not folder_type:
                self.fail('type_not_found', slug=folder_type_slug)
            if not folder_type.folder:
                self.fail('type_not_folder', slug=folder_type_slug)
            data['folder_type'] = folder_type
        else:
            if data['mode'] == DataImportMode.IIIF:
                # folder_type is required in IIIF
                self.fail('iiif_folder_required')
            if not folder_id:
                # Either folder_type or folder_id are required for other imports
                self.fail('folder_required')

        return data


class StartProcessSerializer(serializers.Serializer):
    chunks = serializers.IntegerField(min_value=1, max_value=settings.MAX_CHUNKS, default=1)
    thumbnails = serializers.BooleanField(default=False)
    farm = serializers.PrimaryKeyRelatedField(queryset=Farm.objects.all(), required=False, allow_null=True)
    use_cache = serializers.BooleanField(default=False)
    worker_activity = serializers.BooleanField(default=False)


class ElementsWorkflowSerializer(serializers.Serializer):

    corpus = serializers.UUIDField()
    name = serializers.CharField(required=False, max_length=100)

    # Elements filtering options
    element = serializers.UUIDField(required=False)
    element_name_contains = serializers.CharField(required=False, max_length=250)
    element_type = serializers.SlugField(required=False, max_length=50)
    best_class = BestClassField(required=False)
    selection = serializers.BooleanField(default=False)
    load_children = serializers.BooleanField(default=False)
    use_cache = serializers.BooleanField(default=False)

    def validate_corpus(self, corpus_id):
        try:
            corpus = Corpus.objects.get(id=corpus_id)
        except Corpus.DoesNotExist:
            raise ValidationError(['Corpus with this ID does not exist.'])

        # Assert user has an admin access to the corpus
        access_level = get_max_level(self.context['request'].user, corpus)
        if not access_level:
            raise ValidationError(['Corpus with this ID does not exist.'])
        elif access_level < Role.Admin.value:
            raise ValidationError(['You do not have an admin access to this corpus.'])
        return corpus

    def validate_element(self, element_id):
        if not element_id:
            return None
        try:
            return Element.objects.select_related('type').get(id=element_id)
        except Element.DoesNotExist:
            raise ValidationError(['Element with this ID does not exist.'])

    def validate(self, data):
        """
        Handle filters validation
        """
        corpus = data.get('corpus')

        element = data.get('element')
        elt_type = data.get('element_type')
        best_class = data.get('best_class')
        selection = data.get('selection')

        # Ensure element is a folder or has a zone
        if element and element.type.folder is False and element.zone_id is None:
            raise ValidationError({'element': ['Element has to be a folder or an element with a zone.']})

        if selection and not settings.ARKINDEX_FEATURES['selection']:
            raise ValidationError({'selection': ['Selection is not available on this instance.']})

        if selection and element:
            raise ValidationError({
                'element': ['This field cannot be set with the selection field.'],
                'selection': ['This field cannot be set with the element field.']
            })

        # If element is defined ensure it is part of the right corpus
        if element and element.corpus_id != corpus.id:
            raise ValidationError({
                'element': ['Element is not part of corpus {}'.format(corpus.name)]
            })

        # Check type filter is valid
        if elt_type:
            try:
                data['element_type'] = corpus.types.get(slug=elt_type)
            except ElementType.DoesNotExist:
                raise ValidationError({
                    'element_type': [f'This type does not exist in corpus "{corpus.name}"']
                })

        if best_class is not None and isinstance(best_class, UUID) and not corpus.ml_classes.filter(id=best_class).exists():
            raise ValidationError({
                'best_class': [f'MLClass with ID {best_class} does not exist in corpus "{corpus.name}"']
            })

        return data


class WorkerRunSerializer(serializers.ModelSerializer):
    """
    Serialize a worker run
    """
    parents = serializers.ListField(child=serializers.UUIDField())
    worker_version_id = serializers.UUIDField(source='version_id')
    # Serialize worker with its basic informations
    worker = WorkerLightSerializer(source='version.worker', read_only=True)

    class Meta:
        model = WorkerRun
        read_only_fields = ('id', 'worker', 'dataimport_id')
        fields = (
            'id',
            'parents',
            'worker_version_id',
            'dataimport_id',
            'worker',
        )


class WorkerRunEditSerializer(WorkerRunSerializer):
    """
    Serialize a worker run with only parents as editable field
    """
    worker_version_id = serializers.UUIDField(read_only=True, source='version_id')


class ImportTranskribusSerializer(serializers.Serializer):
    """
    Serialize a Transkribus import
    """
    collection_id = serializers.IntegerField(min_value=1)
    build_entities = serializers.BooleanField(default=False)

    def validate(self, data):
        collection_id = data.get('collection_id')

        # Check Transkribus email
        transkribus_email = self.context['request'].user.transkribus_email
        if not transkribus_email:
            raise serializers.ValidationError(
                {"__all__": ["You have not register your transkribus email"]}
            )

        # Login as Arkindex user on transkribus
        try:
            transkribus_client = TranskribusAPI(email=settings.TRANSKRIBUS_EMAIL, password=settings.TRANSKRIBUS_PASSWORD)
        except Exception:
            raise serializers.ValidationError(
                {"__all__" : [f"Failed to login on Transkribus as {settings.TRANSKRIBUS_EMAIL}"]}
            )

        # Check Arkindex's right
        try:
            users = transkribus_client.list_user_collection(collection_id)
        except Exception:
            raise serializers.ValidationError(
                {"collection_id" : [f"User {settings.TRANSKRIBUS_EMAIL} is not a member of the collection {collection_id}"]}
            )

        # Check user's right
        user = next(filter(lambda user: user["email"] == transkribus_email, users), None)
        if not user:
            raise serializers.ValidationError(
                {"collection_id": [f"User {transkribus_email} is not a member of the collection {collection_id}"]}
            )

        return data


class CreateImportTranskribusErrorResponseSerializer(serializers.Serializer):
    collection_id = serializers.CharField(required=False, help_text="Errors that occured during collection ID field validation.")


class ProcessElementLightSerializer(serializers.ModelSerializer):
    """
    Serialises an Element, using optimized query for ListProcessElement
    """
    type = serializers.CharField(source='type__slug')

    class Meta:
        model = Element
        fields = (
            'id',
            'type',
            'name',
        )
        read_only_fields = fields


class ProcessElementSerializer(ProcessElementLightSerializer):
    image_id = serializers.UUIDField(source='zone__image_id', allow_null=True)
    image_width = serializers.IntegerField(min_value=0, source='zone__image__width', allow_null=True)
    image_height = serializers.IntegerField(min_value=0, source='zone__image__height', allow_null=True)
    image_url = serializers.URLField(allow_null=True)
    polygon = LinearRingField(source='zone__polygon', allow_null=True)

    class Meta(ProcessElementLightSerializer.Meta):
        fields = ProcessElementLightSerializer.Meta.fields + (
            'image_id',
            'image_width',
            'image_height',
            'image_url',
            'polygon',
        )
        read_only_fields = fields
