
import importlib
import uuid
from unittest.mock import MagicMock, patch

import yaml
from django.conf import settings
from django.test import override_settings
from django.urls import reverse
from rest_framework import status
from rest_framework.exceptions import ValidationError

from arkindex.dataimport.models import ActivityState, DataImport, DataImportMode, RepositoryType
from arkindex.dataimport.utils import get_default_farm_id
from arkindex.documents.models import Corpus, ElementType
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role, User
from ponos.models import Farm, State, Task, Workflow

RECIPE = '''
tasks:
  first:
    image: alpine
  second:
    parents:
      - first
    image: hello-world
'''


class TestImports(FixtureAPITestCase):
    """
    Test data imports management
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.creds = cls.user.credentials.get()
        cls.repo = cls.creds.repos.get(type=RepositoryType.IIIF)
        cls.rev = cls.repo.revisions.get()
        cls.img_df = cls.corpus.files.create(
            name='test.jpg',
            size=42,
            content_type='image/jpg',
        )
        cls.pdf_df = cls.corpus.files.create(
            name='test.pdf',
            size=42,
            content_type='application/pdf',
        )
        cls.iiif_df = cls.corpus.files.create(
            name='test.json',
            size=42,
            content_type='application/json',
        )
        cls.page_type = ElementType.objects.get(corpus=cls.corpus, slug='page')

    def setUp(self):
        # Create multiple processes the user can access
        self.user2 = User.objects.create_user('user2@test.test', display_name='Process creator')
        # Guest access (A user own a process on a corpus they are not a member anymore)
        self.user_img_process = DataImport.objects.create(
            mode=DataImportMode.Images,
            creator=self.user,
            corpus=Corpus.objects.create(name='Private'),
        )
        # Contributor access
        self.repo.memberships.create(user=self.user, level=Role.Contributor.value)
        self.repository_process = DataImport.objects.create(
            mode=DataImportMode.Repository,
            creator=self.user2,
            revision=self.rev,
        )
        # Admin access
        self.elts_process = self.corpus.imports.create(mode=DataImportMode.Workers, creator=self.user2)
        self.processes = DataImport.objects.filter(
            id__in=[self.user_img_process.id, self.repository_process.id, self.elts_process.id]
        )

    def _serialize_process(self, process):
        return {
            'name': process.name,
            'id': str(process.id),
            'state': process.state.value,
            'mode': process.mode.value,
            'corpus': process.corpus_id and str(process.corpus.id),
            'workflow': process.workflow and f'http://testserver/ponos/v1/workflow/{process.workflow.id}/',
        }

    def build_task(self, workflow_id, run, state, depth=1):
        """
        Create a simple task with its workflow and process if required
        """
        workflow, created = Workflow.objects.get_or_create(
            id=workflow_id,
            farm=Farm.objects.get(name="Wheat farm"),
            defaults={'recipe': RECIPE}
        )
        if created:
            # Creates the associated dataimport
            DataImport.objects.create(workflow=workflow, mode=DataImportMode.Workers, creator=self.user)
        return Task.objects.create(run=run, depth=depth, workflow=workflow, state=state, slug=str(uuid.uuid4()))

    def test_list_requires_login(self):
        response = self.client.get(reverse('api:import-list'))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_list(self):
        """
        A user can list processes they have access to
        """
        # Order processes
        p1 = self.processes.get(mode=DataImportMode.Workers)
        p1.start()
        p2, p3 = self.processes.exclude(mode=DataImportMode.Workers)
        p2.save()
        # An update on a p1 process task will make it first
        task = p1.workflow.tasks.first()
        task.save()

        self.client.force_login(self.user)
        response = self.client.get(reverse('api:import-list'))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        # Process are ordered
        self.assertDictEqual(response.json(), {
            'count': 3,
            'number': 1,
            'next': None,
            'previous': None,
            'results': [
                self._serialize_process(p)
                for p in [p1, p2, p3]
            ]
        })

    def test_list_with_workflow(self):
        """
        Filter processes that have a workflow i.e. that have been started
        """
        self.user_img_process.start()
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.get(reverse('api:import-list'), {'with_workflow': 'true'})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 1)
        results = data['results']
        self.assertListEqual(results, [{
            'name': None,
            'id': str(self.user_img_process.id),
            'state': State.Unscheduled.value,
            'mode': DataImportMode.Images.value,
            'corpus': str(self.user_img_process.corpus.id),
            'workflow': f'http://testserver/ponos/v1/workflow/{self.user_img_process.workflow.id}/',
        }])

    def test_list_exclude_workflow(self):
        """
        Filter process that have not been started yet
        """
        self.repository_process.start()
        self.elts_process.start()
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.get(reverse('api:import-list'), {'with_workflow': 'false'})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 1)
        results = data['results']
        self.assertListEqual(results, [{
            'name': None,
            'id': str(self.user_img_process.id),
            'state': State.Unscheduled.value,
            'mode': DataImportMode.Images.value,
            'corpus': str(self.user_img_process.corpus.id),
            'workflow': None,
        }])

    def test_list_filter_corpus(self):
        self.client.force_login(self.superuser)
        corpus2 = Corpus.objects.create(name='Another corpus', description='something')
        dataimport2 = corpus2.imports.create(creator=self.user, mode=DataImportMode.Images)

        response = self.client.get(reverse('api:import-list'), {'corpus': str(corpus2.id), 'with_workflow': False})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 1)
        self.assertEqual(data['results'][0]['id'], str(dataimport2.id))

    def test_list_filter_corpus_invalid(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:import-list'), {'corpus': 'oh-no'})
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'corpus': ["'oh-no' is not a valid UUID"]})

    def test_list_filter_mode(self):
        self.client.force_login(self.user)
        dataimport2 = self.corpus.imports.create(
            creator=self.user,
            mode=DataImportMode.PDF,
        )
        response = self.client.get(reverse('api:import-list'), {'mode': DataImportMode.PDF.value, 'with_workflow': False})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 1)
        self.assertEqual(data['results'][0]['id'], str(dataimport2.id))

    def test_list_filter_unexisting_mode(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:import-list'), {'mode': 'unexisting_mode'})
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'mode': ["Mode 'unexisting_mode' does not exist"]})

    def test_list_filter_created(self):
        """
        Display processes that have been created by the user
        """
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:import-list'), {'created': 'true'})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['count'], 1)
        self.assertEqual(data['results'][0]['id'], str(self.user_img_process.id))

    def test_list_filter_not_created(self):
        """
        Display processes that have not been created by the user
        """
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:import-list'), {'created': 'false'})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['count'], 2)
        self.assertCountEqual(
            [p['id'] for p in data['results']],
            [str(p.id) for p in self.processes.exclude(creator=self.user)]
        )

    def test_list_filter_id(self):
        self.client.force_login(self.user)

        # Generate a UUID whose first characters are not the same as the first DataImport
        dataimport_id = str(uuid.uuid4())
        while dataimport_id[:10] == str(self.user_img_process.id)[:10]:
            dataimport_id = str(uuid.uuid4())

        dataimport2 = self.corpus.imports.create(
            id=dataimport_id,
            creator=self.user,
            mode=DataImportMode.Images,
        )
        response = self.client.get(reverse('api:import-list'), {'id': dataimport_id[:10], 'with_workflow': False})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 1)
        self.assertEqual(data['results'][0]['id'], str(dataimport2.id))

    def test_list_no_tasks(self):
        """
        Ensure the DataImport reports an Unscheduled state when there are no tasks in its workflow
        """
        self.assertIsNone(self.user_img_process.workflow)
        self.user_img_process.start()
        self.user_img_process.workflow.tasks.all().delete()

        self.client.force_login(self.user)
        response = self.client.get(reverse('api:import-list'), {'id': str(self.user_img_process.id)})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 1)
        self.assertEqual(data['results'][0]['state'], State.Unscheduled.value)

    def test_list_process_filter_wrong_state_value(self):
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:import-list'), {'state': 'spain'})
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'state': ["State 'spain' does not exist"]})

    def test_list_process_state_filter(self):
        """
        Only last run should be used to filter processes by state
        """
        workflows = [uuid.uuid4() for i in range(3)]
        for task in (
            (workflows[0], 0, State.Failed),
            (workflows[0], 1, State.Completed),

            (workflows[1], 0, State.Error),
            (workflows[1], 0, State.Completed),

            (workflows[2], 0, State.Completed),
            (workflows[2], 1, State.Completed),
            (workflows[2], 2, State.Stopped),
        ):
            self.build_task(*task)
        processes = [DataImport.objects.get(workflow_id=w) for w in workflows]
        # Only first dataimport is completed
        self.assertListEqual([p for p in processes if p.state.value == 'completed'], [processes[0]])

        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.get(reverse('api:import-list'), {'state': 'completed'})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertCountEqual([r['id'] for r in response.json()['results']], [str(processes[0].id)])

    def test_list_process_state_filter_stopped(self):
        workflows = [uuid.uuid4() for i in range(4)]
        for task in (
            # Should be stopping
            (workflows[0], 0, State.Error),
            (workflows[0], 1, State.Stopping),
            # Should be in error state
            (workflows[1], 0, State.Error),
            (workflows[1], 0, State.Stopping),
            # Should be stopped
            (workflows[2], 0, State.Unscheduled),
            (workflows[2], 0, State.Stopped),
            # Should be stopped
            (workflows[3], 0, State.Failed),
            (workflows[3], 1, State.Pending),
            (workflows[3], 1, State.Stopped),
        ):
            self.build_task(*task)
        processes = [DataImport.objects.get(workflow_id=w) for w in workflows]
        self.assertListEqual([p for p in processes if p.state.value == 'stopped'], [processes[2], processes[3]])

        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.get(reverse('api:import-list'), {'state': 'stopped'})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertCountEqual(
            [r['id'] for r in response.json()['results']],
            [str(processes[2].id), str(processes[3].id)]
        )

    def test_list_process_state_filter_unscheduled(self):
        """
        Processes with no task should be interpreted as unscheduled
        """
        workflow_id = uuid.uuid4()
        for task in (
            (workflow_id, 0, State.Completed),
            (workflow_id, 1, State.Unscheduled)
        ):
            self.build_task(*task)
        processes = [
            DataImport.objects.get(workflow_id=workflow_id),
            *self.processes
        ]
        # Workflows with no tasks are considered unscheduled
        self.assertListEqual([p for p in processes if p.state.value == 'unscheduled'], processes)

        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.get(reverse('api:import-list'), {'state': 'unscheduled', 'with_workflow': True})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertCountEqual([r['id'] for r in response.json()['results']], [str(processes[0].id)])

        self.client.force_login(self.user)
        with self.assertNumQueries(5):
            response = self.client.get(reverse('api:import-list'), {'state': 'unscheduled', 'with_workflow': False})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()['count'], len(self.processes))
        self.assertCountEqual([r['id'] for r in response.json()['results']], [str(p.id) for p in processes[1:]])

    def test_details_requires_login(self):
        response = self.client.get(reverse('api:import-details', kwargs={'pk': self.user_img_process.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'Authentication credentials were not provided.'})

    def test_details_no_tasks(self):
        """
        Ensure the DataImport reports an Unscheduled state when there are no tasks in its workflow
        """
        self.assertIsNone(self.user_img_process.workflow)
        self.user_img_process.start()
        self.user_img_process.workflow.tasks.all().delete()
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.get(reverse('api:import-details', kwargs={'pk': self.user_img_process.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['state'], State.Unscheduled.value)

    def test_details(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.get(reverse('api:import-details', kwargs={'pk': self.user_img_process.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data['id'], str(self.user_img_process.id))

    def test_delete_requires_verified(self):
        user = User.objects.create(email='AAAAAAA')
        self.client.force_login(user)
        response = self.client.delete(reverse('api:import-details', kwargs={'pk': self.user_img_process.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You do not have permission to perform this action.'})

    def test_delete_running(self):
        """
        It is not possible to delete a runnin import
        """
        self.client.force_login(self.user)
        self.elts_process.start()
        self.elts_process.workflow.tasks.update(state=State.Running)
        response = self.client.delete(reverse('api:import-details', kwargs={'pk': self.elts_process.id}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'__all__': ['Cannot delete a workflow while it is running']})

    def test_delete_repository_import_no_permission(self):
        """
        Deletion of a repository import requires to be admin on the repository
        """
        self.client.force_login(self.user)
        response = self.client.delete(reverse('api:import-details', kwargs={'pk': self.repository_process.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You do not have a sufficient access level to this process.'})

    def test_delete_corpus_import_no_permission(self):
        """
        A user cannot delete a dataimport linked to a corpus he has no admin access to
        """
        self.client.force_login(self.user)
        self.assertFalse(self.user_img_process.corpus.memberships.filter(user=self.user).exists())
        response = self.client.delete(reverse('api:import-details', kwargs={'pk': self.user_img_process.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You do not have a sufficient access level to this process.'})

    def test_delete_corpus_import_admin(self):
        """
        A user is allowed to delete a dataimport if he has an admin right to its corpus
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(10):
            response = self.client.delete(reverse('api:import-details', kwargs={'pk': self.elts_process.id}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

    def test_delete_superuser(self):
        """
        A superuser is allowed to delete any dataimport
        """
        self.client.force_login(self.superuser)
        with self.assertNumQueries(7):
            response = self.client.delete(reverse('api:import-details', kwargs={'pk': self.user_img_process.id}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

    def test_update_process_requires_login(self):
        response = self.client.delete(reverse('api:import-details', kwargs={'pk': self.user_img_process.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'Authentication credentials were not provided.'})

    def test_update_process_requires_verified(self):
        user = User.objects.create(email="unchecked")
        self.client.force_login(user)
        response = self.client.delete(reverse('api:import-details', kwargs={'pk': self.user_img_process.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You do not have permission to perform this action.'})

    def test_update_process_running(self):
        """
        A running process cannot be updated... (see following test)
        """
        self.client.force_login(self.user)
        self.elts_process.start()
        self.elts_process.workflow.tasks.update(state=State.Running)
        with self.assertNumQueries(7):
            response = self.client.patch(reverse('api:import-details', kwargs={'pk': self.elts_process.id}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'__all__': ['Cannot edit a workflow while it is running']})

    def test_update_process_running_only_name(self):
        """
        Only name attribute of a running process can be patched
        """
        self.client.force_login(self.user)
        self.elts_process.start()
        self.elts_process.workflow.tasks.update(state=State.Running)
        self.assertEqual(self.elts_process.name, None)
        with self.assertNumQueries(9):
            response = self.client.patch(
                reverse('api:import-details', kwargs={'pk': self.elts_process.id}),
                {'name': 'newName'},
                format='json'
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.elts_process.refresh_from_db()
        self.assertEqual(self.elts_process.name, 'newName')

    def test_update_process_no_permission(self):
        """
        A user cannot update a dataimport linked to a corpus he has no admin access to
        """
        self.client.force_login(self.user)
        self.corpus.memberships.filter(user=self.user).update(level=Role.Guest.value)
        response = self.client.patch(reverse('api:import-details', kwargs={'pk': self.elts_process.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You do not have a sufficient access level to this process.'})

    def test_update_process_repository_requires_admin(self):
        """
        Edition of a repository import requires to be admin on the repository
        """
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse('api:import-details', kwargs={'pk': self.repository_process.id}),
            {'name': 'newName'}
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You do not have a sufficient access level to this process.'})

    def test_update_process_wrong_mode(self):
        """
        Other process modes than Workers cannot be patched when there is more than only name attribute
        """
        self.client.force_login(self.user)
        forbidden_modes = list(DataImportMode)
        forbidden_modes.remove(DataImportMode.Workers)
        for mode in forbidden_modes:
            dataimport = DataImport.objects.create(mode=mode, corpus=self.corpus, creator=self.user)
            response = self.client.patch(
                reverse('api:import-details', kwargs={'pk': dataimport.id}),
                {
                    'name': 'newName',
                    'element_name_contains': 'AAA',
                    'element_type': 'page',
                    'load_children': True
                },
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
            self.assertDictEqual(
                response.json(),
                {'__all__': ['Only processes of mode Workers can be updated']}
            )

    def test_update_process_only_name(self):
        """
        Whatever the process mode, name attribute (alone) can be updated
        """
        self.client.force_login(self.user)
        for mode in list(DataImportMode):
            dataimport = DataImport.objects.create(mode=mode, corpus=self.corpus, creator=self.user)
            self.assertEqual(dataimport.name, None)
            response = self.client.patch(
                reverse('api:import-details', kwargs={'pk': dataimport.id}),
                {
                    'name': 'newName'
                },
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
            dataimport.refresh_from_db()
            self.assertEqual(dataimport.name, 'newName')

    def test_update_process_corpus_no_write_right(self):
        self.client.force_login(self.user)
        self.corpus.memberships.filter(user=self.user).update(level=Role.Guest.value)
        response = self.client.patch(reverse('api:import-details', kwargs={'pk': self.elts_process.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You do not have a sufficient access level to this process.'})

    def test_update_process(self):
        """
        Process fields related to elements filtering can be updated
        """
        self.client.force_login(self.user)
        dataimport = DataImport.objects.create(mode=DataImportMode.Workers, corpus=self.corpus, creator=self.user)
        self.assertTupleEqual(
            (dataimport.name, dataimport.name_contains, dataimport.element_type, dataimport.load_children),
            (None, None, None, False)
        )
        with self.assertNumQueries(9):
            response = self.client.patch(
                reverse('api:import-details', kwargs={'pk': dataimport.id}),
                {
                    'name': 'newName',
                    'element_name_contains': 'AAA',
                    'element_type': 'page',
                    'load_children': True,
                },
                format='json'
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        dataimport.refresh_from_db()
        self.assertTupleEqual(
            (dataimport.name, dataimport.name_contains, dataimport.element_type, dataimport.load_children),
            ('newName', 'AAA', self.page_type, True)
        )

    def test_update_process_reset_fields(self):
        self.client.force_login(self.user)
        dataimport = DataImport.objects.create(
            mode=DataImportMode.Workers,
            corpus=self.corpus,
            creator=self.user,
            name='newName',
            name_contains='AAA',
            element_type=self.page_type,
            load_children=True
        )
        response = self.client.patch(
            reverse('api:import-details', kwargs={'pk': dataimport.id}),
            {
                'name': None,
                'element_name_contains': None,
                'element_type': None,
                'load_children': False
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        dataimport.refresh_from_db()
        self.assertTupleEqual(
            (dataimport.name, dataimport.name_contains, dataimport.element_type, dataimport.load_children),
            (None, None, None, False)
        )

    def test_update_process_wrong_type(self):
        self.client.force_login(self.user)
        dataimport = DataImport.objects.create(mode=DataImportMode.Workers, corpus=self.corpus, creator=self.user)
        response = self.client.patch(
            reverse('api:import-details', kwargs={'pk': dataimport.id}),
            {'element_type': 'unexisting_type'},
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(
            response.json(),
            {'element_type': ['Object with slug=unexisting_type does not exist.']}
        )

    def test_update_process_fields(self):
        """
        Ensure determined fields can be updated on a non started workers process.
        Editable fields are `name`, `name_contains`, `element_type`, and `load_children`.
        """
        self.client.force_login(self.user)
        dataimport = DataImport.objects.create(mode=DataImportMode.Workers, corpus=self.corpus, creator=self.user)
        response = self.client.patch(
            reverse('api:import-details', kwargs={'pk': dataimport.id}),
            {
                'id': str(uuid.uuid4()),
                'revision': str(uuid.uuid4()),
                'element': str(uuid.uuid4()),
                'workflow': str(uuid.uuid4()),
                'creator': str(uuid.uuid4()),
                'files': [1, 2, 3],
                'mode': 'iiif',
                'folder_type': 'register',
                'state': 'success',
                'name': 'newName',
                'element_name_contains': 'AAA',
                'element_type': 'page',
                'load_children': True,
                'use_cache': True,
                'activity_state': ActivityState.Pending.value,
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json(), {
            'name': 'newName',
            'element_name_contains': 'AAA',
            'element_type': 'page',
            'load_children': True,
            'use_cache': False,
            'activity_state': ActivityState.Disabled.value,
            'id': str(dataimport.id),
            'corpus': str(self.corpus.id),
            'element': None,
            'files': [],
            'mode': 'workers',
            'revision': None,
            'state': 'unscheduled',
            'workflow': None,
        })

    def test_retry_requires_login(self):
        response = self.client.post(reverse('api:import-retry', kwargs={'pk': self.user_img_process.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_retry_repository_process_no_right(self):
        """
        A user that is not the creator nor admin cannot restart a dataimport that is not linked to any corpus
        """
        self.client.force_login(self.user)
        self.repository_process.start()
        self.repository_process.workflow.tasks.all().update(state=State.Error)
        self.assertEqual(self.repository_process.state, State.Error)
        response = self.client.post(reverse('api:import-retry', kwargs={'pk': self.repository_process.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You do not have an admin access to this process.'})

    def test_retry_only_final(self):
        self.client.force_login(self.user)
        self.elts_process.start()
        self.assertEqual(self.elts_process.state, State.Unscheduled)
        response = self.client.post(reverse('api:import-retry', kwargs={'pk': self.elts_process.id}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_retry_running(self):
        self.client.force_login(self.user)
        self.elts_process.start()
        task = self.elts_process.workflow.tasks.all().last()
        task.state = State.Running
        task.save()
        response = self.client.post(reverse('api:import-retry', kwargs={'pk': self.elts_process.id}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_retry(self):
        self.client.force_login(self.user)
        self.elts_process.start()
        self.elts_process.workflow.tasks.all().update(state=State.Error)
        self.assertEqual(self.elts_process.state, State.Error)
        with self.assertNumQueries(28):
            response = self.client.post(reverse('api:import-retry', kwargs={'pk': self.elts_process.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.elts_process.refresh_from_db()
        self.assertEqual(self.elts_process.state, State.Unscheduled)

    def test_retry_no_workflow(self):
        self.client.force_login(self.user)
        self.assertIsNone(self.elts_process.workflow)
        with self.assertNumQueries(21):
            response = self.client.post(reverse('api:import-retry', kwargs={'pk': self.elts_process.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.elts_process.refresh_from_db()
        self.assertEqual(self.elts_process.state, State.Unscheduled)
        self.assertIsNotNone(self.elts_process.workflow)

    def test_retry_repo_disabled(self):
        self.repository_process.revision = self.rev
        self.repository_process.save()
        self.repository_process.start()
        self.repository_process.workflow.tasks.all().update(state=State.Error)
        self.assertEqual(self.repository_process.state, State.Error)
        self.creds.delete()
        # Allow the user to do the retry
        self.repo.memberships.filter(user=self.user).update(level=Role.Admin.value)
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:import-retry', kwargs={'pk': self.repository_process.id}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), ['Git repository does not have any valid credentials'])

    def test_from_files_requires_login(self):
        response = self.client.post(reverse('api:import-from-files'), {
            'files': [str(self.img_df.id)],
            'folder_type': 'volume',
            'element_type': 'page',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_from_files(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:import-from-files'), {
            'files': [str(self.img_df.id)],
            'folder_type': 'volume',
            'element_type': 'page',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        self.assertEqual(dataimport.mode, DataImportMode.Images)
        self.assertListEqual(list(dataimport.files.all()), [self.img_df])
        self.assertEqual(dataimport.folder_type.slug, 'volume')
        self.assertEqual(dataimport.element_type.slug, 'page')
        self.assertIsNone(dataimport.element)

    def test_from_files_pdf(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:import-from-files'), {
            'files': [str(self.pdf_df.id)],
            'mode': 'pdf',
            'folder_type': 'volume',
            'element_type': 'page',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        self.assertEqual(dataimport.mode, DataImportMode.PDF)
        self.assertListEqual(list(dataimport.files.all()), [self.pdf_df])
        self.assertIsNone(dataimport.element)

    def test_from_files_iiif(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:import-from-files'), {
            'files': [str(self.iiif_df.id)],
            'mode': 'iiif',
            'folder_type': 'volume',
            'element_type': 'page',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        self.assertEqual(dataimport.mode, DataImportMode.IIIF)
        self.assertListEqual(list(dataimport.files.all()), [self.iiif_df])
        self.assertEqual(dataimport.folder_type.slug, 'volume')
        self.assertEqual(dataimport.element_type.slug, 'page')
        self.assertIsNone(dataimport.element)

    def test_from_files_iiif_with_json_charset(self):
        self.iiif_df.content_type = 'application/json;charset=utf-8'
        self.assertEqual(self.iiif_df.content_type, 'application/json;charset=utf-8')

        self.client.force_login(self.user)
        response = self.client.post(reverse('api:import-from-files'), {
            'files': [str(self.iiif_df.id)],
            'mode': 'iiif',
            'folder_type': 'volume',
            'element_type': 'page',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        self.assertEqual(dataimport.mode, DataImportMode.IIIF)
        self.assertListEqual(list(dataimport.files.all()), [self.iiif_df])
        self.assertEqual(dataimport.folder_type.slug, 'volume')
        self.assertEqual(dataimport.element_type.slug, 'page')
        self.assertIsNone(dataimport.element)

    def test_from_files_iiif_with_json_profile_and_charset(self):
        self.iiif_df.content_type = 'application/ld+json;profile="http://iiif.io/api/presentation/2/context.json";charset=utf-8'
        self.assertEqual(self.iiif_df.content_type, 'application/ld+json;profile="http://iiif.io/api/presentation/2/context.json";charset=utf-8')

        self.client.force_login(self.user)
        response = self.client.post(reverse('api:import-from-files'), {
            'files': [str(self.iiif_df.id)],
            'mode': 'iiif',
            'folder_type': 'volume',
            'element_type': 'page',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        self.assertEqual(dataimport.mode, DataImportMode.IIIF)
        self.assertListEqual(list(dataimport.files.all()), [self.iiif_df])
        self.assertEqual(dataimport.folder_type.slug, 'volume')
        self.assertEqual(dataimport.element_type.slug, 'page')
        self.assertIsNone(dataimport.element)

    def test_from_files_iiif_requires_folder_type(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:import-from-files'), {
            'files': [str(self.iiif_df.id)],
            'mode': 'iiif',
            'element_type': 'page',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'non_field_errors': ['IIIF imports require both folder_type and element_type']
        })

    def test_from_files_invalid_mode(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:import-from-files'), {
            'files': [str(self.img_df.id)],
            'folder_type': 'volume',
            'element_type': 'page',
            'mode': 'repository',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_from_files_pdf_wrong_type(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:import-from-files'), {
            'files': [str(self.img_df.id)],
            'mode': 'pdf',
            'folder_type': 'volume',
            'element_type': 'page',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_from_files_images_wrong_type(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:import-from-files'), {
            'files': [str(self.pdf_df.id)],
            'folder_type': 'volume',
            'element_type': 'page',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_from_files_iiif_wrong_type(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:import-from-files'), {
            'files': [str(self.pdf_df.id)],
            'folder_type': 'volume',
            'element_type': 'page',
            'mode': 'iiif',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_from_files_folder_id(self):
        self.client.force_login(self.user)
        volume = self.corpus.elements.get(name='Volume 1')
        response = self.client.post(
            reverse('api:import-from-files'),
            {
                'files': [str(self.pdf_df.id)],
                'mode': 'pdf',
                'folder_id': str(volume.id),
                'element_type': 'page',
            },
            format='json',
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        self.assertEqual(dataimport.element, volume)

    def test_from_files_no_files(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:import-from-files'), {
            'files': [],
            'folder_type': 'volume',
            'element_type': 'page',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_start_process_requires_login(self):
        response = self.client.post(
            reverse('api:process-start', kwargs={'pk': str(self.user_img_process.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_start_process_invalid_dataimport_id(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:process-start', kwargs={'pk': '12341234-1234-1234-1234-123412341234'})
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(response.json(), {'detail': 'Not found.'})

    def test_start_process_wrong_dataimport_mode(self):
        self.assertNotEqual(self.user_img_process.mode, DataImportMode.Workers)
        # grant an admin access to this process
        self.user_img_process.corpus.memberships.create(user=self.user, level=Role.Admin.value)
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:process-start', kwargs={'pk': str(self.user_img_process.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(
            response.json(),
            {'__all__': ['Only a DataImport with Workers mode and not already launched can be started later on']}
        )

    def test_start_process_dataimport_already_started(self):
        dataimport2 = self.corpus.imports.create(creator=self.user, mode=DataImportMode.Workers)
        dataimport2.workflow = Workflow.objects.create(farm_id=get_default_farm_id(), recipe=RECIPE)
        dataimport2.save()
        self.assertIsNotNone(dataimport2.workflow)

        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:process-start', kwargs={'pk': str(dataimport2.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(
            response.json(),
            {'__all__': ['Only a DataImport with Workers mode and not already launched can be started later on']}
        )

    def test_start_process(self):
        """
        A user can start a process with no parameters.
        Defaut chunks, thumbnails and farm are used. Nor cache or workers activity is set.
        """
        dataimport2 = self.corpus.imports.create(creator=self.user, mode=DataImportMode.Workers)
        self.assertIsNone(dataimport2.workflow)

        self.client.force_login(self.user)
        with self.assertNumQueries(21):
            response = self.client.post(
                reverse('api:process-start', kwargs={'pk': str(dataimport2.id)})
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()['id'], str(dataimport2.id))
        dataimport2.refresh_from_db()
        self.assertEqual(dataimport2.state, State.Unscheduled)
        self.assertIsNotNone(dataimport2.workflow)
        # Ensure default parameters are used
        self.assertEqual(dataimport2.workflow.farm_id, get_default_farm_id())
        self.assertEqual(dataimport2.workflow.tasks.count(), 1)
        self.assertIn('--chunks-number 1', dataimport2.workflow.tasks.get().command)

    def test_start_process_select_farm_id(self):
        """
        A user can specify a ponos farm to use for a workflow
        """
        get_default_farm_id()
        barley_farm = Farm.objects.create(name='Barley farm')
        self.assertNotEqual(get_default_farm_id(), barley_farm.id)
        workers_process = self.corpus.imports.create(creator=self.user, mode=DataImportMode.Workers)
        self.client.force_login(self.user)
        with self.assertNumQueries(21):
            response = self.client.post(
                reverse('api:process-start', kwargs={'pk': str(workers_process.id)}),
                {'farm': str(barley_farm.id)}
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        workers_process.refresh_from_db()
        self.assertEqual(workers_process.state, State.Unscheduled)
        self.assertEqual(workers_process.workflow.farm_id, barley_farm.id)

    def test_start_process_wrong_farm_id(self):
        """
        Handle a payload with a wrong farm ID starting a process
        """
        workers_process = self.corpus.imports.create(creator=self.user, mode=DataImportMode.Workers)
        self.client.force_login(self.user)
        wrong_farm_id = uuid.uuid4()
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse('api:process-start', kwargs={'pk': str(workers_process.id)}),
                {'farm': str(wrong_farm_id)}
            )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'farm': [f'Invalid pk "{wrong_farm_id}" - object does not exist.']})

    def test_start_process_params_validation(self):
        dataimport = self.corpus.imports.create(creator=self.user, mode=DataImportMode.Workers)
        self.client.force_login(self.user)
        wrong_params_checks = [
            ({'chunks': 0}, {'chunks': ['Ensure this value is greater than or equal to 1.']}),
            ({'chunks': 20}, {'chunks': ['Ensure this value is less than or equal to 10.']}),
            ({'chunks': 'max'}, {'chunks': ['A valid integer is required.']}),
            ({'thumbnails': 'gloubiboulga'}, {'thumbnails': ['Must be a valid boolean.']})
        ]
        for (params, check) in wrong_params_checks:
            response = self.client.post(reverse('api:process-start', kwargs={'pk': str(dataimport.id)}), params)
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
            self.assertDictEqual(response.json(), check)

    @override_settings(MAX_CHUNKS=42)
    def test_start_process_serializer_chunks_config(self):
        """
        Assert the start process serializer use configuration max chunks
        """
        # Reloading serializer classes is required to update fields attributes from settings
        from arkindex.dataimport.serializers import imports
        imports = importlib.reload(imports)

        params_checks = [
            ({'chunks': 42}, True, None),
            ({'chunks': 43}, False, {'chunks': ['Ensure this value is less than or equal to 42.']}),
        ]
        for (params, is_valid, message) in params_checks:
            serializer = imports.StartProcessSerializer(data=params)
            if not is_valid:
                with self.assertRaisesRegex(ValidationError, 'Ensure this value is less than or equal to 42.'):
                    self.assertEqual(serializer.is_valid(raise_exception=True))
            else:
                self.assertEqual(serializer.is_valid(), True)

    @patch('arkindex.dataimport.models.DataImport.worker_runs')
    def test_start_process_workflow_parameters(self, worker_runs_mock):
        """
        It should be possible to pass chunks and thumbnails parameters when starting a workflow
        """
        dataimport = self.corpus.imports.create(creator=self.user, mode=DataImportMode.Workers)
        # Add a worker run to this process
        run_mock = MagicMock()
        run_mock.version.slug = 'my_worker'
        run_mock.build_task_recipe.return_value = {'image': ''}
        # Cheat to mock a query on process worker runs
        worker_runs_mock.select_related.return_value = worker_runs_mock
        worker_runs_mock.using.return_value = [run_mock]

        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:process-start', kwargs={'pk': str(dataimport.id)}),
            {'chunks': 3, 'thumbnails': 'true'}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        dataimport.refresh_from_db()
        recipe_dump = yaml.safe_load(dataimport.workflow.recipe)

        self.assertEqual(
            list(recipe_dump['tasks'].keys()),
            ['initialisation', 'my_worker_1', 'my_worker_2', 'my_worker_3', 'thumbnails_1', 'thumbnails_2', 'thumbnails_3']
        )

    @patch('arkindex.dataimport.models.DataImport.worker_runs')
    @patch('arkindex.project.triggers.tasks.initialize_activity.delay')
    def test_start_process_options_requires_workers(self, activities_delay_mock, worker_runs_mock):
        """
        Cache and worker activity options can be trigerred hen starting a process
        """
        process = self.corpus.imports.create(
            creator=self.user,
            mode=DataImportMode.Workers,
            element_type=self.corpus.types.get(slug='page')
        )
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse('api:process-start', kwargs={'pk': str(process.id)}),
                {'use_cache': 'true', 'worker_activity': 'true'}
            )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'use_cache': 'The process must should have workers attached to use cached results.',
            'worker_activity': 'The process must have workers attached to handle their activity.'
        })
        process.refresh_from_db()
        self.assertFalse(process.use_cache)
        self.assertEqual(process.activity_state, ActivityState.Disabled)

    @patch('arkindex.dataimport.models.DataImport.worker_runs')
    @patch('arkindex.dataimport.models.DataImport.versions')
    @patch('arkindex.project.triggers.tasks.initialize_activity.delay')
    def test_start_process_options(self, activities_delay_mock, versions_mock, worker_runs_mock):
        """
        Cache and worker activity options requires the process to have workers attached
        """
        process = self.corpus.imports.create(
            creator=self.user,
            mode=DataImportMode.Workers,
            element_type=self.corpus.types.get(slug='page')
        )
        self.assertFalse(process.use_cache)
        self.assertEqual(process.activity_state, ActivityState.Disabled)

        versions_mock.exists.return_value = True

        self.client.force_login(self.user)
        with self.assertNumQueries(20):
            response = self.client.post(
                reverse('api:process-start', kwargs={'pk': str(process.id)}),
                {'use_cache': 'true', 'worker_activity': 'true'}
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        process.refresh_from_db()
        self.assertTrue(process.use_cache)
        # The process activity state is set to Pending while activities are created in an async job
        self.assertEqual(process.activity_state, ActivityState.Pending)
        self.assertEqual(activities_delay_mock.call_count, 1)

    @override_settings(IMPORTS_WORKER_VERSION=uuid.uuid4())
    @override_settings(PONOS_RECIPE={'env': {'ARKINDEX_API_TOKEN': 'testToken'}})
    def test_pdf_import_build_workflow_recipe(self):
        """
        Build a workflow for a PDF import with a worker version defined in settings
        """
        dataimport = self.corpus.imports.create(
            creator=self.user,
            mode=DataImportMode.PDF,
        )
        dataimport.files.create(name='A PDF file', size=10e6, content_type='pdf', corpus=self.corpus)
        workflow = dataimport.build_workflow()

        recipe = yaml.safe_load(workflow.recipe)
        self.assertDictEqual(recipe, {
            'env': {
                'ARKINDEX_API_TOKEN': 'testToken',
                'ARKINDEX_PROCESS_ID': str(dataimport.id),
                'ARKINDEX_CORPUS_ID': str(self.corpus.id),
                'WORKER_VERSION_ID': str(settings.IMPORTS_WORKER_VERSION)
            },
            'tasks': {
                'import': {
                    'command': f'python -m arkindex_tasks.import_files {dataimport.id}',
                    'image': 'registry.gitlab.com/arkindex/tasks'
                }
            }
        })
