from django.conf import settings
from drf_spectacular.utils import extend_schema, extend_schema_view
from elasticsearch_dsl.function import FieldValueFactor
from elasticsearch_dsl.query import FunctionScore, Nested, Q
from rest_framework.generics import ListAPIView

from arkindex.documents.search import search_elements_post, search_entities_post
from arkindex.documents.serializers.search import (
    ElementSearchResultSerializer,
    EntitySearchQuerySerializer,
    EntitySearchResultSerializer,
    SearchQuerySerializer,
)
from arkindex.project.elastic import ESElement, ESEntity
from arkindex.project.mixins import SearchAPIMixin


@extend_schema_view(
    get=extend_schema(
        operation_id='SearchElements',
        tags=['search'],
        parameters=[SearchQuerySerializer],
    )
)
class ElementSearch(SearchAPIMixin, ListAPIView):
    """
    Get a list of elements with their parents, the total number of transcriptions
    in each element, and a few (not all) of their transcriptions, with their source,
    type, zone and image, for a given query.
    """
    serializer_class = ElementSearchResultSerializer

    def get_search(self,
                   corpora_ids=None,
                   query=None,
                   element_type=None,
                   date_lt=None,
                   date_gte=None,
                   min_confidence=0.0):
        assert corpora_ids, 'Must filter by corpora'

        # Note that sorting by parents will not sort properly if there are multiple parents
        # https://www.elastic.co/guide/en/elasticsearch/reference/current/search-request-body.html#_sort_mode_option
        search = ESElement.search() \
            .sort('parents', '_score') \
            .source(fields=['date_range']) \
            .filter('terms', corpus=corpora_ids)

        if date_lt or date_gte:
            date_range = {'relation': 'intersects'}
            if date_lt:
                date_range['lt'] = date_lt
            if date_gte:
                date_range['gte'] = date_gte

            search = search.filter('range', date_range=date_range)

        if element_type:
            search = search.filter('match', type=element_type)

        nested_query = Q('range', transcriptions__score={'gte': min_confidence})
        if query:
            nested_query &= Q('simple_query_string', query=query, fields=['transcriptions.text'])

        elastic_query = Nested(
            path='transcriptions',
            inner_hits={'size': settings.ES_INNER_RESULTS_LIMIT},
            score_mode='sum',
            query=FunctionScore(
                query=nested_query,
                functions=[FieldValueFactor(field='transcriptions.score')],
            ),
        )
        if query:
            elastic_query |= Q('wildcard', references='*{}*'.format(query.lower()))

        search = search.query(elastic_query)
        return search

    def post_process(self, *args, **kwargs):
        return search_elements_post(*args)


@extend_schema_view(
    get=extend_schema(
        operation_id='SearchEntities',
        tags=['search'],
        parameters=[EntitySearchQuerySerializer],
    )
)
class EntitySearch(SearchAPIMixin, ListAPIView):
    serializer_class = EntitySearchResultSerializer
    query_serializer_class = EntitySearchQuerySerializer

    def get_search(self, query=None, type=None, corpora_ids=None):
        assert corpora_ids, 'Must filter by corpora'

        search = ESEntity.search() \
            .filter('terms', corpus=corpora_ids)

        if query:
            search = search.query('simple_query_string', query=query, fields=['name'])
        if type:
            search = search.query('match', type=type.value)

        return search

    def post_process(self, *args, **kwargs):
        return search_entities_post(*args)
