import logging
from typing import Optional

from django.db import transaction
from django.db.models import Q
from django_rq import job
from rq import get_current_job

from arkindex.dataimport.models import ActivityState, DataImport, DataImportElement, WorkerActivity, WorkerRun
from arkindex.documents.indexer import Indexer
from arkindex.documents.managers import ElementQuerySet
from arkindex.documents.models import (
    Classification,
    Corpus,
    Element,
    ElementPath,
    Entity,
    EntityLink,
    MetaData,
    Selection,
    Transcription,
    TranscriptionEntity,
)

logger = logging.getLogger(__name__)


@job
def reindex_start(corpus_id: Optional[str] = None,
                  element_id: Optional[str] = None,
                  entity_id: Optional[str] = None,
                  transcriptions: bool = True,
                  elements: bool = True,
                  entities: bool = True,
                  drop: bool = False) -> None:
    indexer = Indexer()

    if drop:
        if transcriptions:
            indexer.drop_index(Transcription.es_document)
        if elements:
            indexer.drop_index(Element.es_document)
        if entities:
            indexer.drop_index(Entity.es_document)
        indexer.setup()

    if entity_id:
        # Pick entity only
        transcriptions_queryset = Transcription.objects.none()
        elements_queryset = Element.objects.none()
        entities_queryset = Entity.objects.filter(id=entity_id)
    elif element_id or corpus_id:
        if element_id:
            # Pick this element, and all its children
            elements_queryset = Element.objects.filter(Q(id=element_id) | Q(paths__path__contains=[element_id]))
        else:
            # Pick all elements in the corpus
            elements_queryset = Element.objects.filter(corpus_id=corpus_id)

        transcriptions_queryset = Transcription.objects.filter(element__in=elements_queryset)
        entities_queryset = Entity.objects.filter(
            Q(metadatas__element__in=elements_queryset)
            | Q(transcriptions__element__in=elements_queryset)
        )
    else:
        transcriptions_queryset = Transcription.objects.all()
        elements_queryset = Element.objects.all()
        entities_queryset = Entity.objects.all()

    if transcriptions:
        indexer.run_index(transcriptions_queryset.select_related('element'), bulk_size=400)
    if elements:
        indexer.run_index(
            elements_queryset.select_related('type').prefetch_related('metadatas', 'transcriptions'),
            bulk_size=100,
        )
    if entities:
        indexer.run_index(entities_queryset, bulk_size=400)


@job('high')
def corpus_delete(corpus_id: str) -> None:
    # Note that this can be None when the task is run outside of a RQ worker (e.g. unit test)
    rq_job = get_current_job()
    corpus = Corpus.objects.get(id=corpus_id)
    logger.info(f'Deleting {corpus!r}')

    # Delete all related objects, bypassing RESTRICTs deletion rules
    # and Django's way of loading everything into memory before deleting.
    querysets = [
        corpus.allowed_metadatas.all(),
        WorkerRun.objects.filter(dataimport__corpus_id=corpus_id),
        # Using two separate queries is a lot faster than using a OR for large corpora
        DataImportElement.objects.filter(dataimport__corpus_id=corpus_id),
        DataImportElement.objects.filter(element__corpus_id=corpus_id),
        # DataImport-DataFile M2M with implicit model
        DataImport.files.through.objects.filter(dataimport__corpus_id=corpus_id),
        DataImport.files.through.objects.filter(datafile__corpus_id=corpus_id),
        corpus.imports.all(),
        corpus.files.all(),
        MetaData.objects.filter(element__corpus_id=corpus_id),
        EntityLink.objects.filter(role__corpus_id=corpus_id),
        corpus.roles.all(),
        TranscriptionEntity.objects.filter(entity__corpus_id=corpus_id),
        TranscriptionEntity.objects.filter(transcription__element__corpus_id=corpus_id),
        corpus.entities.all(),
        Classification.objects.filter(element__corpus_id=corpus_id),
        corpus.ml_classes.all(),
        Transcription.objects.filter(element__corpus_id=corpus_id),
        ElementPath.objects.filter(element__corpus_id=corpus_id),
        Selection.objects.filter(element__corpus_id=corpus_id),
        corpus.elements.all(),
        corpus.types.all(),
        corpus.memberships.all(),
        Corpus.objects.filter(id=corpus_id),
    ]

    for i, queryset in enumerate(querysets):
        if rq_job:
            rq_job.set_progress(i / len(querysets))
            rq_job.set_description(f'Deleting {queryset.model.__name__} on corpus {corpus.name}')
        deleted_count = queryset._raw_delete(using='default')
        logger.info(f'Deleted {deleted_count} {queryset.model.__name__}')

    logger.info(f'Deleted corpus {corpus_id}')


@job('high')
def element_trash(queryset: ElementQuerySet, delete_children: bool) -> None:
    queryset.trash(delete_children=delete_children)


@job('high')
def worker_results_delete(corpus_id: str, version_id: str, parent_id: str) -> None:
    """
    Recursively delete all Worker Results produced by a specific WorkerVersion on a whole corpus or under
    a specified parent element (parent element included).
    """
    elements = Element.objects.filter(worker_version_id=version_id, corpus_id=corpus_id)
    classifications = Classification.objects.filter(worker_version_id=version_id, element__corpus_id=corpus_id)
    transcriptions = Transcription.objects.filter(worker_version_id=version_id, element__corpus_id=corpus_id)
    transcription_entities = TranscriptionEntity.objects.filter(transcription__worker_version_id=version_id, transcription__element__corpus_id=corpus_id)

    if parent_id:
        elements = elements.filter(Q(id=parent_id) | Q(paths__path__overlap=[parent_id]))
        classifications = classifications.filter(Q(element_id=parent_id) | Q(element__paths__path__overlap=[parent_id]))
        transcriptions = transcriptions.filter(Q(element_id=parent_id) | Q(element__paths__path__overlap=[parent_id]))
        transcription_entities = transcription_entities.filter(Q(transcription__element_id=parent_id) | Q(transcription__element__paths__path__overlap=[parent_id]))

    elements.trash()
    classifications.delete()
    transcription_entities.delete()
    transcriptions._raw_delete(using='default')


@job('default', timeout=3600)
def initialize_activity(process: DataImport):
    """
    List all worker versions used in a process and initialize their activity on processed elements.
    Timeout is set to 1 hour
    """
    try:
        with transaction.atomic():
            for version_id in process.versions.values_list('id', flat=True):
                WorkerActivity.objects.bulk_insert(worker_version_id=version_id, elements_qs=process.list_elements())
    except Exception as e:
        process.activity_state = ActivityState.Error
        process.save()
        raise e
    process.activity_state = ActivityState.Ready
    process.save()
