import uuid

import sqlparse
from django.db.models.sql.constants import LOUTER
from django.urls import reverse
from rest_framework import status

from arkindex.dataimport.models import WorkerVersion
from arkindex.documents.models import Corpus
from arkindex.project.tests import FixtureAPITestCase


class TestListElements(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.worker_version = WorkerVersion.objects.get(worker__slug='reco')

    def test_list_elements_sql(self):
        """
        Test ListElements with assertExactQueries.

        ListElements prefetches zone IDs using a separate query that gets zones, images and servers all at once.
        Django gathers zone IDs from elements in an unordered manner, which means that we cannot put an ordered list
        of zone IDs easily in the SQL fixture.

        To overcome this, this test uses sqlparse to look through the AST for the prefetch query and find the list of
        zones, then builds a list of UUIDs and compares it against an expected list of zone IDs.
        It then sends the actual list of UUIDs as a formatting parameter to assertExactQueries, allowing the fixture
        file to just say "WHERE id IN ({zone_param})" and not break on this filter.

        Future maintainers: If you want to dig into the query, the simplest way is to explore the result of
        sqlparse.parse in an interactive shell.  You could copy-paste an SQL string from the file in sql_validation,
        or run pdb straight from this test:

        zone_query = sql_parse.parse(…)
        import pdb
        pdb.set_trace()

        Then use `p zone_query` to see the query, `p zone_query.tokens` to see its child tokens, etc.
        You can access a zone ID directly using `p zone_query.tokens[-1].tokens[-1].tokens[-1].tokens[1].tokens[0].value`.
        Learn more about sqlparse tokens here: https://sqlparse.readthedocs.io/en/latest/analyzing/
        """
        params = {'corpus_id': self.corpus.id}

        expected_zone_ids = set(
            self.corpus.elements.all()
            .order_by('corpus', 'type__slug', 'name', 'id')[:20]
            .values_list('zone_id', flat=True)
        )

        with self.assertExactQueries('list_elements.sql', params=params) as ctx:
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

            # Look for the zone prefetch query
            # sqlparse.parse returns a tuple of statements, we will only have one
            zone_query = sqlparse.parse(next(query['sql'] for query in ctx if 'images_zone' in query['sql']))[0]
            # Get the where clause
            where = next(token for token in zone_query.tokens if isinstance(token, sqlparse.sql.Where))
            # Get the IN comparison in the where clause
            comparison = next(token for token in where.tokens if isinstance(token, sqlparse.sql.Comparison))
            # Find the parentheses
            parens = next(token for token in comparison.tokens if isinstance(token, sqlparse.sql.Parenthesis))
            # Find the identifier list
            id_list = next(token for token in parens.tokens if isinstance(token, sqlparse.sql.IdentifierList))

            # Parse the identifier list, handling both UUIDs and NULL.
            actual_zone_ids = set()
            for id_token in id_list.tokens:
                # In this case, the Identifier is a string parsed as a UUID: it is made of three Tokens.
                # The first token is the UUID as a string, the second is the cast operator (::) and the third is the UUID type.
                # We grab the string, remove its quotes, and parse it as a Python UUID.
                if isinstance(id_token, sqlparse.sql.Identifier):
                    actual_zone_ids.add(uuid.UUID(id_token.tokens[0].value.strip("'")))
                # NULL is parsed as a single Token class with a Keyword ttype.
                elif id_token.match(sqlparse.tokens.Keyword, "NULL"):
                    actual_zone_ids.add(None)

            # Test the zone IDs separately.
            self.assertSetEqual(expected_zone_ids, actual_zone_ids, 'Unexpected prefetched zone IDs')

            # Now that we tested the zone IDs, we can put them in the SQL fixture and make assertExactQueries
            # 'ignore' them when we exit the context manager.
            params['zone_param'] = str(parens)

    def test_list_elements_filter_name(self):
        nameSelected = 'ol'
        with self.assertNumQueries(4):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                data={'name': nameSelected}
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()['results']
        for result in results:
            self.assertIn(nameSelected, result['name'])

    def test_list_elements_corpus_uuid4(self):
        """
        Ensure the corpus filter allows any UUID, not just version 4 UUIDs.
        Using the version=x argument on uuid.UUID causes the version and variant bits
        of a UUID to be changed on instanciation; since we generate v4 RFC 4122-compliant UUIDs,
        this would work in most cases - but we could have any UUID version or variant,
        generated by some other tool, which would cause strange 404 errors.
        """
        str_id = '12345678-1234-5678-1234-567812345678'
        self.assertNotEqual(
            uuid.UUID(str_id),
            uuid.UUID(str_id, version=4),
        )
        corpus = Corpus.objects.create(id=uuid.UUID(str_id), name='A corpus', public=True)
        element_type = corpus.types.create(slug='element', display_name='Generic element')
        element = corpus.elements.create(type=element_type, name='Something')

        response = self.client.get(reverse('api:corpus-elements', kwargs={'corpus': str_id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 1)
        self.assertEqual(data['results'][0]['id'], str(element.id))
        self.assertEqual(data['results'][0]['corpus']['id'], str_id)

    def test_list_elements_filter_type(self):
        response = self.client.get(
            reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
            data={'type': 'act'}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()['results']
        for result in results:
            self.assertEqual(result['type'], 'act')

    def test_list_with_has_children(self):
        with self.assertNumQueries(6):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                data={'with_has_children': True}
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()

        self.assertDictEqual(
            {element['name']: element['has_children'] for element in data['results']},
            {
                'Act 1': True,
                'Act 2': True,
                'Act 3': True,
                'Act 4': True,
                'Act 5': True,
                'Text line': False,
                'Surface A': False,
                'Surface B': False,
                'Surface C': False,
                'Surface D': False,
                'Surface E': False,
                'Surface F': False,
                'Volume 1': True,
                'Volume 1, page 1r': True,
                'Volume 1, page 1v': True,
                'Volume 1, page 2r': True,
                'Volume 2': True,
                'Volume 2, page 1r': False,
                'Volume 2, page 1v': False,
                'Volume 2, page 2r': False,
            }
        )

    def test_list_top_level_left_join(self):
        """
        Ensure the top_level option on ListElements triggers a left join.
        """
        query = self.corpus.elements.filter(paths__isnull=True, name='something').query
        # Ensure the documents_elementpath table is joined via a LEFT JOIN
        self.assertEqual(query.alias_map['documents_elementpath'].join_type, LOUTER)
        # Ensure the IS NULL goes last
        self.assertTrue(str(query).endswith('AND "documents_elementpath"."id" IS NULL)'))

    def test_list_top_level(self):
        with self.assertNumQueries(4):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                data={'top_level': True},
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [element['name'] for element in response.json()['results']],
            ['Volume 1', 'Volume 2'],
        )

    def test_list_top_level_default_type(self):
        """
        Use the corpus top level type to filter top level elements when available
        """
        # Add a top level element of type page
        element_type = self.corpus.types.create(slug='top_level', display_name='Some type')
        self.corpus.elements.create(type=element_type, name='Generic element')
        self.corpus.top_level_type = element_type
        self.corpus.save()

        with self.assertNumQueries(4):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                data={'top_level': True},
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [element['name'] for element in response.json()['results']],
            ['Generic element'],
        )

    def test_list_elements_filter_worker_version(self):
        self.corpus.elements.filter(name__contains='Volume 2').update(worker_version=self.worker_version)

        with self.assertNumQueries(5):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                data={'worker_version': str(self.worker_version.id)},
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [element['name'] for element in response.json()['results']],
            ['Volume 2, page 1r', 'Volume 2, page 1v', 'Volume 2, page 2r', 'Volume 2'],
        )

    def test_list_elements_filter_worker_version_validation(self):
        with self.assertNumQueries(1):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                data={'worker_version': 'blah'},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'worker_version': ['Invalid UUID']})

        with self.assertNumQueries(2):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                data={'worker_version': uuid.uuid4()},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'worker_version': ['This worker version does not exist.']})

    def test_list_elements_filter_modified_since_not_modified(self):
        with self.assertNumQueries(2):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                HTTP_IF_MODIFIED_SINCE='Thu, 02 Apr 2099 13:37:42 GMT',
            )
        self.assertEqual(response.status_code, status.HTTP_304_NOT_MODIFIED)

    def test_list_elements_filter_modified_since_bad_format(self):
        with self.assertNumQueries(1):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                HTTP_IF_MODIFIED_SINCE='AAAAAAAAAAAAAAAAA',
            )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'If-Modified-Since': ['Bad date format']})

    def test_list_elements_filter_modified_since(self):
        self.corpus.elements.filter(name='Volume 1').update(updated='2099-04-02T13:37:43Z')
        with self.assertNumQueries(5):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                HTTP_IF_MODIFIED_SINCE='Thu, 02 Apr 2099 13:37:42 GMT',
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()['count'], self.corpus.elements.count())

    def test_list_elements_with_corpus_false(self):
        with self.assertNumQueries(4):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id})
                + '?with_corpus=false'
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertSetEqual(
            set(element.get('corpus') for element in response.json()['results']),
            {None}
        )
