from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus, ElementType
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import User


class TestElementType(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.element_type = ElementType.objects.create(corpus=cls.corpus, slug='element', display_name='Element')
        cls.user = User.objects.get(email="user@user.fr")
        cls.private_corpus = Corpus.objects.create(name='private')

    def test_create_element_type_requires_login(self):
        response = self.client.post(reverse('api:element-type-create'), {
            'corpus': self.corpus.id,
            'slug': 'New_element_type',
            'display_name': 'New element type',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_create_element_type_no_access(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:element-type-create'), {
            'corpus': self.private_corpus.id,
            'slug': 'New_element_type',
            'display_name': 'New element type',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_create_element_type_wrong_slug(self):
        self.client.force_login(self.superuser)
        response = self.client.post(reverse('api:element-type-create'), {
            'corpus': self.corpus.id,
            'slug': self.element_type.slug,
            'display_name': 'New element type',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_create_element_type(self):
        self.client.force_login(self.superuser)
        response = self.client.post(reverse('api:element-type-create'), {
            'corpus': self.corpus.id,
            'slug': 'New_element_type',
            'display_name': 'New element type',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        self.assertIn('id', data)
        element_type = ElementType.objects.get(id=data['id'])
        self.assertDictEqual(
            data,
            {
                'id': str(element_type.id),
                'corpus': str(element_type.corpus.id),
                'slug': element_type.slug,
                'display_name': element_type.display_name,
                'folder': element_type.folder
            }
        )

    def test_update_element_type_requires_login(self):
        response = self.client.post(reverse('api:element-type', kwargs={'pk': self.element_type.id}), {
            'slug': 'New_element_type',
            'display_name': 'New element type',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_update_element_type_no_access(self):
        private_element_type = ElementType.objects.create(corpus=self.private_corpus, slug='element', display_name='Element')
        self.client.force_login(self.user)
        response = self.client.patch(reverse('api:element-type', kwargs={'pk': private_element_type.id}), {
            'slug': 'New_element_type',
            'display_name': 'New element type',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_update_element_type_wrong_slug(self):
        element_type = ElementType.objects.create(corpus=self.corpus, slug='new_type', display_name='Element')
        self.client.force_login(self.superuser)
        response = self.client.patch(reverse('api:element-type', kwargs={'pk': self.element_type.id}), {
            'slug': element_type.slug,
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'slug': ['Slug must be unique']})

    def test_update_element_type(self):
        self.client.force_login(self.superuser)
        response = self.client.patch(reverse('api:element-type', kwargs={'pk': self.element_type.id}), {
            'slug': 'New_element_type',
            'display_name': 'New element type',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.element_type.refresh_from_db()
        self.assertIn('id', data)
        self.assertDictEqual(
            data,
            {
                'id': str(self.element_type.id),
                'slug': self.element_type.slug,
                'display_name': self.element_type.display_name,
                'folder': self.element_type.folder
            }
        )
