from django.conf import settings
from django.db import models
from django.utils.functional import cached_property

# Lower-case markers that *may* indicate a IIIF server handles
# images after that path
IIIF_MARKERS = ('/iiif/', '/i3f/')


class ImageServerManager(models.Manager):

    @cached_property
    def local(self):
        return self.get(id=settings.LOCAL_IMAGESERVER_ID)

    def from_url(self, url):
        """
        Find a matching ImageServer from an image URL. Requires the server to be validated.
        May raise ImageServer.DoesNotExist or ImageServer.MultipleObjectsReturned.
        """
        servers = self \
            .annotate(image_url=models.Value(url, output_field=models.CharField())) \
            .filter(validated=True, image_url__startswith=models.F('url'))

        servers_count = servers.count()
        if not servers_count:

            # Try to detect a IIIF marker in the url
            # and create a new image server automatically
            for marker in IIIF_MARKERS:
                if marker not in url.lower():
                    continue
                server_url = url[:url.lower().index(marker) + len(marker)]
                return self.create(
                    url=server_url,
                    display_name=server_url,
                    validated=True,
                )

            raise self.model.DoesNotExist(
                'This resource does not match any existing image server'
            )
        elif servers_count > 1:
            raise self.model.MultipleObjectsReturned(
                'This resource matched multiple existing image servers'
            )

        return servers.get()
