from rest_framework import permissions
from rest_framework.exceptions import PermissionDenied


class AllowNone(object):
    """
    Systematically refuse permission
    """
    def has_permission(self, request, view):
        return None


class InternalGroupPermissionMixin(object):
    """
    Immediately authenticate any non-admin Internal users.
    This allows admin-like access to Ponos tasks while preventing actual access to the Django admin.
    """

    def has_permission(self, request, view):
        if request.user.is_authenticated and \
                not request.user.is_admin and \
                request.user.is_internal:
            return True

        return super().has_permission(request, view)


class UserScopePermissionMixin(object):
    """
    When the `scopes` attribute is defined, ensure an authenticated user has said scopes,
    no matter which HTTP method is used.
    When the `<method>_scopes` attribute is defined for a given HTTP method, ensures the user
    has said scopes only for this HTTP method.
    """

    def get_scopes(self, view, method):
        specific_scopes_attr = '{}_scopes'.format(method.lower())
        scopes = list(getattr(view, 'scopes', []))
        scopes.extend(getattr(view, specific_scopes_attr, []))
        return scopes

    def has_permission(self, request, view):
        scopes = self.get_scopes(view, request.method)
        # Skip this permission mixin if there are no defined scopes or if the user is an admin
        if not scopes or (request.user.is_authenticated and request.user.is_admin):
            return super().has_permission(request, view)

        missing_scopes = set(scopes)
        if request.user.is_authenticated:
            # Tell which scopes are missing by subtracting found scopes from the expected scopes
            missing_scopes -= set(request.user.user_scopes.filter(scope__in=scopes).values_list('scope', flat=True))

        if not missing_scopes:
            return super().has_permission(request, view)

        # For "read-only" methods, let the rest of the permission decide when the scopes are not matched
        # This allows -OrReadOnly permissions to work
        if request.method in permissions.SAFE_METHODS and super().has_permission(request, view):
            return True

        # If scopes are not matched and permissions did not allow the read-only method, cause a custom error
        raise PermissionDenied({
            'detail': PermissionDenied.default_detail,
            'missing_scopes': [scope.value for scope in missing_scopes],
        })


class VerifiedEmailPermissionMixin(object):
    """
    Only allow read-only requests (aka safe methods: HEAD, GET, OPTIONS)
    for non-admin users with unverified emails.
    """

    def has_permission(self, request, view):
        if request.user.is_authenticated \
                and not request.user.is_admin \
                and not request.user.verified_email:
            # If the user has an unverified email, only allow HEAD, GET and OPTIONS.
            return super().has_permission(request, view) and request.method in permissions.SAFE_METHODS

        return super().has_permission(request, view)


class AllowAny(InternalGroupPermissionMixin, permissions.AllowAny):
    pass


class IsAuthenticated(InternalGroupPermissionMixin,
                      UserScopePermissionMixin,
                      permissions.IsAuthenticated):
    pass


class IsAuthenticatedOrReadOnly(InternalGroupPermissionMixin,
                                UserScopePermissionMixin,
                                permissions.IsAuthenticatedOrReadOnly):
    pass


class IsAdminUser(InternalGroupPermissionMixin,
                  UserScopePermissionMixin,
                  permissions.IsAdminUser):
    pass


class IsInternalUser(InternalGroupPermissionMixin, AllowNone):
    pass


class IsVerified(VerifiedEmailPermissionMixin, IsAuthenticated):
    pass


class IsVerifiedOrReadOnly(VerifiedEmailPermissionMixin, IsAuthenticatedOrReadOnly):
    pass
