import uuid

from django.contrib.auth.models import AnonymousUser
from django.contrib.contenttypes.models import ContentType

from arkindex.dataimport.models import DataImport, DataImportMode, Repository, RepositoryType, Revision
from arkindex.documents.models import Corpus
from arkindex.project.mixins import ACLMixin, CorpusACLMixin, ProcessACLMixin, RepositoryACLMixin, WorkerACLMixin
from arkindex.project.tests import FixtureTestCase
from arkindex.users.models import Group, Right, Role, User
from arkindex.users.utils import filter_rights, get_max_level


class TestACLMixin(FixtureTestCase):

    @classmethod
    def setUpClass(cls):
        r"""
        Create user and groups with rights on a Corpus and a Repository
        We use a simple rights configuration for those tests

           User1      User2     User3
          /  |  \      / |     /   |
         |   |  100   10 |   100   |
        50  100   \  /   90   |    75
         |   |   Group1  |  Group2 |
         | Worker / \    |   /\    |
         |   |   80  75  | 100 50  |
          \  |  /      \ | /    \  |
           Repo1     Corpus1  Corpus2
        """

        super().setUpClass()
        cls.user1 = User.objects.create_user('user1@test.test', display_name='User1')
        cls.user2 = User.objects.create_user('user2@test.test', display_name='User2')
        cls.user3 = User.objects.create_user('user3@test.test', display_name='User3')

        cls.group1 = Group.objects.create(name='Group1')
        cls.group2 = Group.objects.create(name='Group2')

        cls.repo1 = Repository.objects.create(type=RepositoryType.Worker, url='http://repo1')
        cls.worker = cls.repo1.workers.create(name='repo1 worker', slug='worker')

        cls.corpus1 = Corpus.objects.create(name="Corpus1", id=uuid.UUID('bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb'))
        cls.corpus2 = Corpus.objects.create(name="Corpus2", id=uuid.UUID('aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa'))

        Right.objects.bulk_create([
            Right(user=cls.user1, content_object=cls.group1, level=100),
            Right(user=cls.user1, content_object=cls.repo1, level=50),
            Right(user=cls.user1, content_object=cls.worker, level=100),
            Right(user=cls.user2, content_object=cls.group1, level=10),
            Right(user=cls.user2, content_object=cls.corpus1, level=90),
            Right(user=cls.user3, content_object=cls.group2, level=100),
            Right(user=cls.user3, content_object=cls.corpus2, level=75),
            Right(group=cls.group1, content_object=cls.repo1, level=80),
            Right(group=cls.group1, content_object=cls.corpus1, level=75),
            Right(group=cls.group2, content_object=cls.corpus1, level=100),
            Right(group=cls.group2, content_object=cls.corpus2, level=50),
        ])
        cls.corpus_type = ContentType.objects.get_for_model(Corpus)
        cls.repo_type = ContentType.objects.get_for_model(Repository)
        cls.group_type = ContentType.objects.get_for_model(Group)

    def test_filter_right(self):
        # List all the rights with a privilege greater than 80 for user 3
        params = {
            'user_id': self.user3.id,
            'group_id': self.group2.id,
            'group_type_id': self.group_type.id,
            'corpus_type_id': self.corpus_type.id,
            'level': 80,
        }
        with self.assertExactQueries('corpus_rights_filter.sql', params=params):
            # List to queryset to fire the DB request
            corpora = list(filter_rights(self.user3, Corpus, 80))
        self.assertCountEqual(corpora, [self.corpus1])

    def test_filter_right_including_public(self):
        # Listing readable objects involve listing public objects with a SQL Union
        params = {
            'user_id': self.user1.id,
            'group_id': self.group2.id,
            'group_type_id': self.group_type.id,
            'corpus_type_id': self.corpus_type.id,
            'level': Role.Guest.value,
        }
        with self.assertExactQueries('corpus_rights_filter_public.sql', params=params):
            corpora = list(filter_rights(self.user1, Corpus, Role.Guest.value))
        self.assertCountEqual(
            list(corpora),
            [self.corpus1, self.corpus]
        )

    def test_right_direct_access(self):
        # User 2 has a direct access to the above corpus with a level of 90
        acl_mixin = ACLMixin(user=self.user2)
        params = {
            'user_id': self.user2.id,
            'group_id': self.group2.id,
            'corpus_id': self.corpus1.id,
            'group_type_id': self.group_type.id,
            'corpus_type_id': self.corpus_type.id,
            'level': 80,
        }
        with self.assertExactQueries('corpus_right_access.sql', params=params):
            has_access = acl_mixin.has_access(self.corpus1, 80)
        self.assertTrue(has_access)

    def test_right_group_members_restriction(self):
        # User rights on corpora via a group are restricted to user level inside the group
        acl_mixin = ACLMixin(user=self.user1)
        self.assertFalse(
            acl_mixin.has_access(self.corpus1, 80)
        )

    def test_corpus_acl_mixin_has_read_access(self):
        corpus_acl_mixin = CorpusACLMixin(user=self.user2)
        with self.assertNumQueries(3):
            read_access = corpus_acl_mixin.has_read_access(self.corpus1)
        self.assertTrue(read_access)

    def test_public_corpus_has_read_access(self):
        # Any user should have a read access to public corpora
        corpus_acl_mixin = CorpusACLMixin(user=AnonymousUser())
        with self.assertNumQueries(0):
            read_access = corpus_acl_mixin.has_read_access(self.corpus)
        self.assertTrue(read_access)

    def test_corpus_acl_mixin_has_write_access(self):
        # User2 has a direct access to Corpus1 with an adequate level
        corpus_acl_mixin = CorpusACLMixin(user=self.user2)
        with self.assertNumQueries(3):
            write_access = corpus_acl_mixin.has_write_access(self.corpus1)
        self.assertTrue(write_access)

    def test_corpus_acl_mixin_has_admin_access(self):
        # Admin access requires either to have an admin right or to be a django admin/internal user
        admin_access = [
            (self.user, 3, False),
            (self.superuser, 0, True),
            (self.user1, 3, False),
            (self.user2, 3, False),
            (self.user3, 3, True)
        ]
        for user, queries, access_check in admin_access:
            corpus_acl_mixin = CorpusACLMixin(user=user)
            with self.assertNumQueries(queries):
                admin_access = corpus_acl_mixin.has_admin_access(self.corpus1)
            self.assertEqual(admin_access, access_check)
            ContentType.objects.clear_cache()

    def test_corpus_acl_mixin_writable(self):
        corpus_acl_mixin = CorpusACLMixin(user=self.user1)
        with self.assertNumQueries(3):
            corpora = list(corpus_acl_mixin.writable_corpora)
        self.assertCountEqual(
            list(corpora),
            [self.corpus1]
        )

    def test_repo_acl_mixin_has_read_access(self):
        repo_acl_mixin = RepositoryACLMixin(user=self.user2)
        with self.assertNumQueries(3):
            read_access = repo_acl_mixin.has_read_access(self.repo1)
        self.assertTrue(read_access)

    def test_repo_acl_mixin_has_write_access(self):
        repo_acl_mixin = RepositoryACLMixin(user=self.user2)
        with self.assertNumQueries(3):
            exec_access = repo_acl_mixin.has_execution_access(self.repo1)
        self.assertFalse(exec_access)

    def test_repo_acl_mixin_has_admin_access(self):
        # Admin access requires either to have an admin right or to be a django admin/internal user
        admin_access = [
            (self.user, 3, False),
            (self.superuser, 0, True),
            (self.user1, 3, False),
            (self.user2, 3, False),
            (self.user3, 3, False)
        ]
        for user, queries, access_check in admin_access:
            repo_acl_mixin = RepositoryACLMixin(user=user)
            with self.assertNumQueries(queries):
                admin_access = repo_acl_mixin.has_admin_access(self.repo1)
            self.assertEqual(admin_access, access_check)
            ContentType.objects.clear_cache()

    def test_repo_acl_mixin_readable(self):
        repo_acl_mixin = RepositoryACLMixin(user=self.user2)
        with self.assertNumQueries(3):
            repos = list(repo_acl_mixin.readable_repositories)
        self.assertCountEqual(
            list(repos),
            [self.repo1]
        )

    def test_repo_acl_mixin_executable(self):
        repo_acl_mixin = RepositoryACLMixin(user=self.user2)
        with self.assertNumQueries(3):
            repos = list(repo_acl_mixin.executable_repositories)
        self.assertEqual(repos, [])

    def test_worker_acl_mixin_has_worker_access(self):
        # Worker access may be determined via its repository

        group1_admin = User.objects.create_user('group1+admin@test.test')
        self.group1.memberships.create(user=group1_admin, level=Role.Admin.value)
        access_cases = [
            (self.superuser, 100, 0, True),
            (self.user1, 100, 3, True),
            (self.user, 10, 5, False),
            (self.user2, 80, 5, False),
            (self.user3, 10, 5, False),
            (self.user2, 10, 5, True),
            (group1_admin, 100, 5, False),
            (group1_admin, 80, 5, True),
        ]
        for user, level, queries, access_check in access_cases:
            worker_acl_mixin = WorkerACLMixin(user=user)
            with self.assertNumQueries(queries):
                access = worker_acl_mixin.has_worker_access(self.worker, level)
            self.assertEqual(access, access_check)
            ContentType.objects.clear_cache()

    def test_corpus_readable_orderable(self):
        # Assert corpora retrieved via the mixin are still orderable
        corpus_acl_mixin = CorpusACLMixin(user=self.user3)
        with self.assertNumQueries(3):
            corpora = list(corpus_acl_mixin.readable_corpora.order_by('name'))
        self.assertListEqual(
            [c.name for c in corpora],
            ['Corpus1', 'Corpus2', 'Unit Tests']
        )

    def test_super_admin_readable_corpora(self):
        # A super admin should retrieve all existing corpora with Admin rigths
        corpus_acl_mixin = CorpusACLMixin(user=self.superuser)
        with self.assertNumQueries(1):
            corpora = list(corpus_acl_mixin.readable_corpora)
        self.assertCountEqual(
            list(corpora),
            list(Corpus.objects.all())
        )

    def test_anonymous_user_readable_corpora(self):
        # An anonymous user should have guest access to any public corpora
        corpus_acl_mixin = CorpusACLMixin(user=AnonymousUser())
        with self.assertNumQueries(1):
            corpora = list(corpus_acl_mixin.readable_corpora)
        self.assertCountEqual(
            list(corpora),
            list(Corpus.objects.filter(public=True))
        )

    def test_corpus_right_and_public(self):
        # User specific rigths should be returned instead of the the defaults access for public rigths
        Right.objects.create(user=self.user3, content_object=self.corpus, level=42)
        corpus_acl_mixin = CorpusACLMixin(user=self.user3)
        with self.assertNumQueries(2):
            corpora = list(corpus_acl_mixin.readable_corpora)
        self.assertCountEqual(
            list(corpora),
            [self.corpus1, self.corpus2, self.corpus]
        )

    def test_max_level_does_not_exists(self):
        with self.assertNumQueries(3):
            self.assertEqual(get_max_level(self.user1, self.corpus2), None)

    def test_max_level_user1(self):
        with self.assertNumQueries(3):
            self.assertEqual(get_max_level(self.user1, self.repo1), 80)

    def test_max_level_user2(self):
        with self.assertNumQueries(3):
            self.assertEqual(get_max_level(self.user2, self.repo1), 10)

    def test_max_level_user3(self):
        with self.assertNumQueries(3):
            self.assertEqual(get_max_level(self.user3, self.corpus2), 75)

    def test_max_level_public(self):
        with self.assertNumQueries(3):
            self.assertEqual(get_max_level(self.user1, self.corpus), Role.Guest.value)

    def test_process_access_creator(self):
        """
        A process creator always has a read access
        """
        process = DataImport.objects.create(mode=DataImportMode.Workers, creator=self.user1, corpus=self.corpus2)
        with self.assertNumQueries(3):
            self.assertEqual(ProcessACLMixin(user=self.user1).process_access_level(process), Role.Guest.value)

    def test_process_access_project(self):
        """
        A project attached to a process defines its access rights
        """
        process = DataImport.objects.create(mode=DataImportMode.Workers, creator=self.user2, corpus=self.corpus1)
        with self.assertNumQueries(3):
            self.assertEqual(ProcessACLMixin(user=self.user1).process_access_level(process), 75)

    def test_process_access_docker_build(self):
        """
        Access to a docker build process is possible for someone with an access to the origin repository
        """
        process = DataImport.objects.create(
            mode=DataImportMode.Repository,
            creator=self.user2,
            revision=self.repo1.revisions.create(hash='42', message='Message', author='User 1')
        )
        with self.assertNumQueries(3):
            self.assertEqual(ProcessACLMixin(user=self.user1).process_access_level(process), 80)

    def test_process_access_list(self):
        """
        A user can list process either if they are creator, or have rights on its corresponding project or repository
        """
        # Readable processes
        process_1 = DataImport.objects.create(mode=DataImportMode.PDF, creator=self.user1)
        process_2 = DataImport.objects.create(mode=DataImportMode.Workers, creator=self.user2, corpus=self.corpus1)
        process_3 = DataImport.objects.create(
            mode=DataImportMode.Repository,
            creator=self.user2,
            revision=self.repo1.revisions.create(hash='42', message='Message', author='User 1')
        )
        # Non readable processes
        DataImport.objects.create(mode=DataImportMode.Workers, creator=self.user2)
        DataImport.objects.create(mode=DataImportMode.Workers, creator=self.user2, corpus=self.corpus2)
        DataImport.objects.create(
            mode=DataImportMode.Repository,
            creator=self.user2,
            revision=Revision.objects.get(message='My w0rk3r')
        )

        with self.assertNumQueries(4):
            readable_process_ids = list(
                ProcessACLMixin(user=self.user1).readable_processes.values_list('id', flat=True)
            )
        self.assertCountEqual(
            readable_process_ids,
            [process_1.id, process_2.id, process_3.id]
        )
