import datetime
import urllib.parse
from abc import ABC, abstractmethod

import requests
from django.conf import settings
from django.urls import reverse
from gitlab import Gitlab, GitlabError
from rest_framework.exceptions import NotAuthenticated


class OAuthProvider(ABC):
    """
    An OAuth authentication provider.
    """

    display_name = ""
    slug = ""

    def __init__(self, request=None, credentials=None, url=None):
        if request is not None:
            # Allow Django and Django REST Framework requests
            from django.http.request import HttpRequest
            from rest_framework.request import Request
            assert isinstance(request, (HttpRequest, Request))
        if credentials is not None:
            from arkindex.users.models import OAuthCredentials
            assert isinstance(credentials, OAuthCredentials)
            self.url = credentials.provider_url
        elif url is not None:
            self.url = url
        self.request = request
        self.credentials = credentials

    @classmethod
    @abstractmethod
    def enabled(cls):
        """
        Boolean stating the provider's availability.
        """

    @abstractmethod
    def get_callback_uri(self):
        """
        Get the OAuth callback URI
        """

    @abstractmethod
    def get_authorize_uri(self):
        """
        Get the OAuth authorization endpoint URI
        """

    @abstractmethod
    def handle_callback(self):
        """
        Handle a OAuth callback and save token data. Should raise exceptions if the process fails.
        """

    @abstractmethod
    def disconnect(self):
        """
        Erase token data and logout the user from the service.
        """


class GitLabOAuthProvider(OAuthProvider):

    display_name = 'GitLab'
    slug = 'gitlab'
    url = 'https://gitlab.com'
    authorize_endpoint = '/oauth/authorize'
    token_endpoint = '/oauth/token'

    @classmethod
    def enabled(cls):
        return settings.GITLAB_APP_ID and settings.GITLAB_APP_SECRET

    def get_callback_uri(self):
        if not self.request:
            return

        url = reverse('api:oauth-callback', kwargs={'provider': self.slug})
        if settings.BACKEND_PUBLIC_URL_OAUTH:
            return urllib.parse.urljoin(settings.BACKEND_PUBLIC_URL_OAUTH, url)

        return self.request.build_absolute_uri(url)

    def get_authorize_uri(self):
        if not self.request or not self.credentials:
            return
        return '{}?{}'.format(
            urllib.parse.urljoin(self.credentials.provider_url, self.authorize_endpoint),
            urllib.parse.urlencode({
                'client_id': settings.GITLAB_APP_ID,
                'redirect_uri': self.get_callback_uri(),
                'scope': 'api',
                'response_type': 'code',
                'state': str(self.credentials.id),
            }),
        )

    def handle_callback(self):
        if not self.request:
            return

        state = self.request.GET.get('state')
        if not state:
            raise ValueError('No state hash')

        self.credentials = self.request.user.credentials.get(id=state)

        if not any(param in self.request.GET for param in ('code', 'error')):
            raise ValueError('Callback called without a valid response')
        if 'error' in self.request.GET:
            raise ValueError(self.request.GET.get('error_description', self.request.GET['error']))

        from arkindex.users.models import OAuthStatus
        assert self.credentials.status != OAuthStatus.Done, 'Cannot overwrite existing credentials'

        response = requests.post(
            urllib.parse.urljoin(self.credentials.provider_url, self.token_endpoint),
            {
                'client_id': settings.GITLAB_APP_ID,
                'client_secret': settings.GITLAB_APP_SECRET,
                'code': self.request.GET.get('code', ''),
                'grant_type': 'authorization_code',
                'redirect_uri': self.get_callback_uri(),
            }
        )
        response.raise_for_status()
        data = response.json()

        assert 'access_token' in data, 'GitLab returned no OAuth token'
        assert data.get('token_type', '').lower() == 'bearer', 'Unsupported OAuth token type'

        self.credentials.token = data['access_token']
        self.credentials.refresh_token = data.get('refresh_token')
        if 'expires_in' in data:
            self.credentials.expiry = datetime.datetime.now() + datetime.timedelta(0, int(data['expires_in']))

        gl = Gitlab(self.credentials.provider_url, oauth_token=self.credentials.token)
        gl.auth()
        self.credentials.account_name = gl.user.username

        self.credentials.status = OAuthStatus.Done
        self.credentials.save()

    def disconnect(self):
        if not self.request and self.credentials:
            raise NotAuthenticated()

        if not self.credentials.token or not self.credentials.repos.exists():
            return

        # Remove all webhooks
        try:
            gl = Gitlab(self.url, oauth_token=self.credentials.token)
            for repo in self.credentials.repos.all():
                project = gl.projects.get(urllib.parse.urlsplit(repo.url).path.strip('/'))
                hook_url = self.request.build_absolute_uri(
                    reverse('api:import-hook', kwargs={'pk': repo.id})
                )
                # Try to find the webhook
                hook = next((h for h in project.hooks.list() if h.url == hook_url), None)
                if hook:
                    hook.delete()
        except GitlabError:
            pass


oauth_providers = [
    GitLabOAuthProvider,
]


def get_provider(name):
    return next(filter(lambda p: p.slug == name, oauth_providers), None)
